from __future__ import annotations
import itertools
import math
from typing import Dict, List, Mapping, Set, Tuple, Union
import warnings

import demes
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize

__all__ = [
    "get_fig_axes",
    "size_max",
    "size_min",
    "log_size_heuristic",
    "log_time_heuristic",
    "separation_heuristic",
    "minimal_crossing_positions",
    "optimise_positions",
]


# Override the symbols that are returned when calling dir(<module-name>).
def __dir__():
    return sorted(__all__)


# A colour is either a colour name string (e.g. "blue"), or an RGB triple,
# or an RGBA triple.
Colour = Union[str, Tuple[float, float, float], Tuple[float, float, float, float]]
# A mapping from a string to a colour, or just a single colour
ColourOrColourMapping = Union[Mapping[str, Colour], Colour]


def _get_times(graph: demes.Graph) -> Set[float]:
    """
    Get all times in the graph.

    :param demes.Graph: The graph.
    :return: The set of times found in the graph.
    :rtype: set[float]
    """
    times = set()
    for deme in graph.demes:
        times.add(deme.start_time)
        for epoch in deme.epochs:
            times.add(epoch.end_time)
    for migration in graph.migrations:
        times.add(migration.start_time)
        times.add(migration.end_time)
    for pulse in graph.pulses:
        times.add(pulse.time)
    return times


def _inf_start_time(graph: demes.Graph, inf_ratio: float, log_scale: bool) -> float:
    """
    Calculate the value on the time axis that will be used instead of infinity.

    :param demes.Graph graph: The graph.
    :param float inf_ratio: The proportion of the time axis that will be used
        for the time interval which stretches towards infinity.
    :param bool log_scale: The time axis uses a log scale.
    :return: The time.
    :rtype: float
    """
    times = _get_times(graph)
    times.discard(math.inf)
    oldest_noninf_time = max(times)

    if oldest_noninf_time == 0:
        # All demes are root demes and have a constant size.
        # About 100 generations is a nice time scale to draw, and we use 113
        # specifically so that the infinity line extends slightly beyond the
        # last tick mark that matplotlib autogenerates.
        inf_start_time: float = 113
    else:
        if log_scale:
            if oldest_noninf_time <= 1:
                # A log scale is a terrible choice for this graph.
                warnings.warn(
                    "Graph contains features at 0 < time <= 1, which will not "
                    "be visible on a log scale."
                )
                oldest_noninf_time = 2
            inf_start_time = math.exp(math.log(oldest_noninf_time) / (1 - inf_ratio))
        else:
            inf_start_time = oldest_noninf_time / (1 - inf_ratio)
    return inf_start_time


def _get_colours(
    graph: demes.Graph,
    colours: ColourOrColourMapping | None = None,
    default_colour: Colour = "gray",
) -> Mapping[str, Colour]:
    """
    Convert the polymorphic ``colours`` into a dictionary of colours,
    keyed by deme name.

    :param demes.Graph graph: The graph to which colours will apply.
    :param colours: The colour or colours.
        * If ``colours`` is ``None``, the default colour map will be used.
        * If ``colours`` is a dict, it must map deme names to colours.
          All demes not in the dict will be drawn with ``default_colour``.
        * Otherwise, if ``colours`` can be interpreted as a matplotlib
          colour, all demes will be drawn with this colour.
    :type: dict or str
    """
    new_colours: Dict[str, Colour]  # Stop mypy from complaining.
    if colours is None:
        if len(graph.demes) <= 10:
            cmap = plt.get_cmap("tab10")
        elif len(graph.demes) <= 20:
            cmap = plt.get_cmap("tab20")
        else:
            raise ValueError(
                "Graph has more than 20 demes, so colours must be specified."
            )
        new_colours = {deme.name: cmap(j) for j, deme in enumerate(graph.demes)}
    elif isinstance(colours, Mapping):
        bad_names = list(colours.keys() - set([deme.name for deme in graph.demes]))
        if len(bad_names) > 0:
            raise ValueError(
                f"Colours given for deme(s) {bad_names}, but deme(s) were "
                "not found in the graph."
            )
        new_colours = {deme.name: default_colour for deme in graph.demes}
        new_colours.update(colours)
    else:
        # Try to interpret as a matplotlib colour.
        try:
            colour = matplotlib.colors.to_rgba(colours)
        except ValueError as e:
            raise ValueError(
                f"Colour '{colours}' not interpretable as a matplotlib colour"
            ) from e
        new_colours = {deme.name: colour for deme in graph.demes}
    return new_colours


def get_fig_axes(
    aspect: float | None = None, scale: float | None = None, **kwargs
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """
    Get a matplotlib figure and axes.

    The default width and height of a matplotlib figure is 6.4 x 4.8 inches.
    To create axes on a figure with other sizes, the ``figsize`` argument
    can be passed to :func:`matplotlib.pyplot.subplots`.
    The ``get_fig_axes()`` function creates an axes on a figure with an
    alternative parameterisation that is useful for screen display of
    vector images.

    An ``aspect`` parameter sets the aspect ratio, and a ``scale`` parameter
    multiplies the figure size. Increasing the scale will have the effect of
    decreasing the size of objects in the figure (including fonts),
    and increasing the amount of space between objects.

    :param aspect: The aspect ratio (height/width) of the figure.
        This value will be passed to :func:`matplotlib.figure.figaspect` to
        obtain the figure's width and height dimensions.
        If not specified, 9/16 will be used.
    :param scale: Multiply the figure width and height by this value.
        If not specified, 1.0 will be used.
    :param kwargs: Further keyword args will be passed directly to
        :func:`matplotlib.pyplot.subplots`.
    :return: A 2-tuple containing the matplotlib figure and axes,
        as returned by :func:`matplotlib.pyplot.subplots`.
    :rtype: tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
    """
    if aspect is None:
        aspect = 9.0 / 16.0
    if scale is None:
        scale = 1.0
    fig, ax = plt.subplots(figsize=scale * plt.figaspect(aspect), **kwargs)
    if not fig.get_constrained_layout():
        if hasattr(fig, "set_layout_engine"):
            fig.set_layout_engine("tight")
        else:
            # matplotlib < 3.6 compat
            fig.set_tight_layout(True)
    return fig, ax


def size_max(graph: demes.Graph) -> float:
    """
    Get the maximum deme size in the graph.

    :param demes.Graph graph: The graph.
    :return: The maximum deme size.
    :rtype: float
    """
    return max(
        max(epoch.start_size, epoch.end_size)
        for deme in graph.demes
        for epoch in deme.epochs
    )


def size_min(graph: demes.Graph) -> float:
    """
    Get the minimum deme size in the graph.

    :param demes.Graph graph: The graph.
    :return: The minimum deme size.
    :rtype: float
    """
    return min(
        min(epoch.start_size, epoch.end_size)
        for deme in graph.demes
        for epoch in deme.epochs
    )


def log_size_heuristic(graph: demes.Graph) -> bool:
    """
    Decide whether or not to use log scale for sizes.

    :param demes.Graph graph: The graph.
    :return: True if log scale should be used or False otherwise.
    :rtype: bool
    """
    if size_max(graph) / size_min(graph) > 4:
        log_size = True
    else:
        log_size = False
    return log_size


def log_time_heuristic(graph: demes.Graph) -> bool:
    """
    Decide whether or not to use log scale for times.

    :param demes.Graph graph:
        The graph.
    :return:
        True if log scale should be used or False otherwise.
    """
    times = _get_times(graph)
    times.discard(0)
    times.discard(math.inf)
    if len(times) > 0 and max(times) / min(times) > 4:
        log_time = True
    else:
        log_time = False
    return log_time


def _contemporaries_max(graph: demes.Graph) -> int:
    """The maximum number of contemporary demes at any time."""
    c_max = 1
    for deme_j in graph.demes:
        c = 1
        for deme_k in graph.demes:
            if deme_j is not deme_k and _intersect(deme_j, deme_k):
                c += 1
        if c > c_max:
            c_max = c
    return c_max


def separation_heuristic(graph: demes.Graph) -> float:
    """
    Find a reasonable separation distance for deme positions.

    :param demes.Graph graph:
        The graph.
    :return:
        The separation distance.
    """
    # This looks ok. See spacing.pdf produced by tests/plot_examples.py.
    c = _contemporaries_max(graph)
    return (1 + 0.5 * math.log(c)) * size_max(graph)


def _intersect(
    dm_j: Union[demes.Deme, demes.AsymmetricMigration],
    dm_k: Union[demes.Deme, demes.AsymmetricMigration],
) -> bool:
    """Returns true if dm_j and dm_k intersect in time."""
    return not (dm_j.end_time >= dm_k.start_time or dm_k.end_time >= dm_j.start_time)


def _coexistence_indices(graph: demes.Graph) -> List[Tuple[int, int]]:
    """Pairs of indices of demes that exist simultaneously."""
    contemporaries = []
    for j, deme_j in enumerate(graph.demes):
        for k, deme_k in enumerate(graph.demes[j + 1 :], j + 1):
            if _intersect(deme_j, deme_k):
                contemporaries.append((j, k))
    return contemporaries


def _successors_indices(graph: demes.Graph) -> Dict[int, List[int]]:
    """Graph successors, but use indices rather than deme names"""
    idx = {deme.name: j for j, deme in enumerate(graph.demes)}
    successors = dict()
    for deme, children in graph.successors().items():
        if len(children) > 0:
            successors[idx[deme]] = [idx[child] for child in children]
    return successors


def _interactions_indices(graph: demes.Graph, *, unique: bool) -> List[Tuple[int, int]]:
    """Pairs of indices of demes that exchange migrants (migrations or pulses)."""
    idx = {deme.name: j for j, deme in enumerate(graph.demes)}
    interactions = []
    for migration in graph.migrations:
        interactions.append((idx[migration.source], idx[migration.dest]))
    for pulse in graph.pulses:
        for source in pulse.sources:
            interactions.append((idx[source], idx[pulse.dest]))

    if unique:
        # Remove duplicates.
        i2 = set()
        for a, b in interactions:
            if (b, a) in i2:
                continue
            i2.add((a, b))
        interactions = list(i2)

    return interactions


def _get_line_candidates(graph: demes.Graph, unique: bool):
    """
    Construct candidate orderings that, if present, would cause lines to cross.

    Each candidate is a triplet of demes, [A, B, C], which says that when the
    demes A,B,C are ordered with B between A and C, then there will be a line
    drawn between A and C that will cross deme B.

    :return:
        A 2d numpy array of candidate indices, where each row is a candidate
        triplet, and the array entries are the deme indexes (which correspond
        to the ordering of the demes in the graph).
    """
    candidates = []
    # Ancestry lines.
    for child in graph.demes:
        for parent_name in child.ancestors:
            for deme in graph.demes:
                if deme is child or deme.name == parent_name:
                    continue
                if deme.start_time > child.start_time > deme.end_time:
                    candidates.append((child.name, deme.name, parent_name))
    # Pulse lines.
    for pulse in graph.pulses:
        for source in pulse.sources:
            for deme in graph.demes:
                if deme.name in [pulse.dest, source]:
                    continue
                if deme.start_time > pulse.time > deme.end_time:
                    candidates.append((source, deme.name, pulse.dest))
    # Migration blocks.
    for migration in graph.migrations:
        for deme in graph.demes:
            if deme.name in (migration.dest, migration.source):
                continue
            if _intersect(deme, migration):
                candidates.append((migration.source, deme.name, migration.dest))

    if unique:
        # Remove duplicates.
        c2 = set()
        for a, b, c in candidates:
            if (c, b, a) in c2:
                continue
            c2.add((a, b, c))
        candidates = list(c2)

    idx = {deme.name: j for j, deme in enumerate(graph.demes)}
    return np.array([(idx[a], idx[b], idx[c]) for a, b, c in candidates])


def _line_crossings(xx: np.ndarray, candidates: np.ndarray) -> np.ndarray:
    """
    Count the number of lines that cross one or more demes.

    :param xx:
        The positions of the demes in the graph. This is a 2d array,
        where each row in xx is a distinct vector of proposed positions.
        Specifically, xx[j, k] is the position of deme k in the j'th
        proposed positions vector.
    :param candidates:
        A 2d array of candidate indices. See _get_line_candidates().
    :return:
        The number of lines crossing demes. This is a 1d array
        of counts, one for each vector of proposed positions.
    """
    if candidates.size == 0:
        return np.array(0)
    a = xx[..., candidates[:, 0]]
    b = xx[..., candidates[:, 1]]
    c = xx[..., candidates[:, 2]]
    return np.logical_or(
        np.logical_and(a < b, b < c), np.logical_and(a > b, b > c)
    ).sum(axis=-1)


def minimal_crossing_positions(
    graph: demes.Graph,
    *,
    sep: float,
    unique_interactions: bool,
    maxiter: int = 1_000_000,
    seed: int = 1234,
) -> Dict[str, float]:
    """
    Find an ordering of demes that minimises lines crossing demes.

    Lines may be for ancestry, pulses, or migrations. Finding the optimal
    ordering is hard, but counting how many lines cross for any given ordering
    is simple, and can be calculated for multiple candidate orderings using
    vectorised numpy operations. So a naive algorithm is used to search for
    a good ordering:

      - if :math:`n!` <= ``maxiter``, where ``n`` is the number of demes in the
        graph, then all possible orderings may be evaluated,
      - otherwise up to ``maxiter`` random orderings are evaluated.

    If a proposed ordering produces zero line crossings, the algorithm
    terminates early. Furthermore, if no ordering is superior to the order
    that demes were specified in the model (i.e. the order in ``graph.demes``),
    then the model's order will be used.

    :param graph:
        Graph for which positions should be obtained.
    :param sep:
        The separation distance between demes.
    :param maxiter:
        Maximum number of orderings to search through.
    :param seed:
        Seed for the random number generator.
    :return:
        A dictionary mapping deme names to positions.
    """
    # Initial ordering according to the model.
    x0 = np.arange(0, sep * len(graph.demes), sep)

    def propose_positions_batches(num_proposals: int, batch_size=200):
        n = len(graph.demes)
        if math.factorial(n) <= num_proposals:
            # Generate all permutations.
            iperms = itertools.permutations(x0, n)
            while True:
                x_batch = np.array(list(itertools.islice(iperms, batch_size)))
                if x_batch.size == 0:
                    break
                yield x_batch
        else:
            # Generate random permutations.
            rng = np.random.default_rng(seed)
            remaining = num_proposals
            while remaining > 0:
                batch_size = min(batch_size, remaining)
                yield rng.permuted(
                    np.tile(x0, batch_size).reshape(batch_size, -1), axis=1
                )
                remaining -= batch_size

    candidates = _get_line_candidates(graph, unique=unique_interactions)
    crosses = _line_crossings(x0, candidates)
    x_best = x0
    if crosses > 0:
        for x_proposal_batch in propose_positions_batches(maxiter):
            proposal_crosses = _line_crossings(x_proposal_batch, candidates)
            best = np.argmin(proposal_crosses)
            if proposal_crosses[best] < crosses:
                crosses = proposal_crosses[best]
                x_best = x_proposal_batch[best]
                if crosses == 0:
                    break

    return {deme.name: float(pos) for deme, pos in zip(graph.demes, x_best)}


class _PositionsObjective:
    """
    Helper class encapsulating the quadratic objective for
    :func:`optimise_positions`, its gradient and Hessian.
    """

    def __init__(
        self, successors: Dict[int, List[int]], interactions: List[Tuple[int, int]]
    ):
        self._successors = successors
        self._interactions = interactions
        # Hessian matrix: computed lazily on first request.
        self._H = None

    def f(self, x):
        """
        Return the sum of the squared distances between interacting demes.
        This is the minimisation objective for :func:`optimise_positions`.
        """
        val = 0
        for parent, children in self._successors.items():
            a = x[parent]
            b = np.mean([x[child] for child in children])
            val += (a - b) ** 2
        for j, k in self._interactions:
            val += (x[j] - x[k]) ** 2
        return val

    def g(self, x):
        """
        Return the Jacobian (partial derivatives) of :func:`f`.
        """
        g = np.zeros_like(x)
        for parent, children in self._successors.items():
            a = x[parent]
            b = np.mean([x[child] for child in children])
            g[parent] += 2 * (a - b)
            m = len(children)
            for child in children:
                g[child] += 2 * (b - a) / m
        for j, k in self._interactions:
            g[j] += 2 * (x[j] - x[k])
            g[k] += 2 * (x[k] - x[j])
        return g

    def h(self, x):
        """
        Return the Hessian (matrix of second partial derivatives) of :func:`f`.
        This is constant with respect to x, so we compute it only once and
        cache the result.
        """
        if self._H is None:
            # Compute the Hessian.
            n = len(x)
            H = np.zeros((n, n), dtype=float)
            for parent, children in self._successors.items():
                m = len(children)
                H[parent, parent] += 2.0
                for child in children:
                    H[parent, child] += -2.0 / m
                    H[child, parent] += -2.0 / m
                contrib = 2.0 / (m * m)
                for s in children:
                    for t in children:
                        H[s, t] += contrib
            for j, k in self._interactions:
                H[j, j] += 2.0
                H[k, k] += 2.0
                H[j, k] += -2.0
                H[k, j] += -2.0
            self._H = H
        return self._H


def optimise_positions(
    graph: demes.Graph,
    *,
    positions: Mapping[str, float],
    sep: float,
    unique_interactions: bool,
) -> Dict[str, float]:
    """
    Optimise the given positions into a tree-like layout.

    The objective is to minimise the distances:
      - from each parent deme to the mean position of its children,
      - and between interacting demes (where interactions are either
        migrations or pulses).

    Subject to the constraints that contemporaneous demes:
      - are ordered like in the input ``positions``,
      - and have a minimum separation distance ``sep``.

    :param graph:
        Graph for which positions should be optimised.
    :param positions:
        A dictionary mapping deme names to positions.
    :param sep:
        The minimum separation distance between contemporary demes.
    :return:
        A dictionary mapping deme names to positions.
    """
    contemporaries = _coexistence_indices(graph)
    successors = _successors_indices(graph)
    interactions = _interactions_indices(graph, unique=unique_interactions)
    if len(contemporaries) == 0:
        # There are no constraints, so stack demes on top of each other.
        return {deme.name: 0 for deme in graph.demes}
    if len(successors) == 0 and len(interactions) == 0:
        # Nothing to optimise. Use the positions provided to us.
        return dict(positions)

    x0 = np.array([positions[deme.name] for deme in graph.demes])
    # Place the first deme at position 0.
    x0 -= x0[0]

    # Build matrix of linear constraints to ensure that contemporaries
    # are ordered and separated by sep.
    C = []
    for j, k in contemporaries:
        if x0[j] < x0[k]:
            j, k = k, j
        c = [0] * len(x0)
        c[j] = 1
        c[k] = -1
        C.append(c)
    constraints = [scipy.optimize.LinearConstraint(C, lb=sep, ub=np.inf)]

    obj = _PositionsObjective(successors, interactions)

    res = scipy.optimize.minimize(
        obj.f,
        x0,
        jac=obj.g,
        hess=obj.h,
        method="trust-constr",
        constraints=constraints,
        bounds=scipy.optimize.Bounds(lb=np.min(x0) - sep, ub=np.max(x0) + sep),
    )

    if not res.success:
        warnings.warn(
            f"Failed to optimise: {res}\n\n"
            "Please report this in the issue tracker at "
            "https://github.com/grahamgower/demesdraw/issues"
        )
        return dict(positions)

    x = res.x

    return {graph.demes[j].name: float(xj) for j, xj in enumerate(x)}



import pathlib
import functools
import warnings

import demes
import matplotlib

matplotlib.use("Agg")


@functools.lru_cache(maxsize=None)
def example_files():
    cwd = pathlib.Path(__file__).parent.resolve()
    example_dir = cwd / "../examples"
    files = list(example_dir.glob("**/*.yaml"))
    assert len(files) > 1
    return files


@functools.lru_cache(maxsize=None)
def example_graphs():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "Multiple pulses", UserWarning, "demes")
        graphs = [demes.load(fn) for fn in example_files()]
    return graphs


@functools.lru_cache(maxsize=None)
def example_demes():
    return [deme for graph in example_graphs() for deme in graph.demes]



import sys

import packaging.version


def is_new_stable(old_stable, tag):
    """
    Return 0 if `tag` should be the new 'stable' docs folder, 1 otherwise.
    """
    try:
        v_t = packaging.version.Version(tag)
    except packaging.version.InvalidVersion:
        return 1
    v_s = packaging.version.Version(old_stable)
    return 0 if not v_t.is_prerelease and v_t > v_s else 1


if __name__ == "__main__":
    assert is_new_stable("0.2.0", "taggy-mac-tag-tag") != 0
    assert is_new_stable("0.2.0", "0.1.0") != 0
    assert is_new_stable("0.2.0", "0.1.0b1") != 0
    assert is_new_stable("0.2.0", "0.2.0b1") != 0
    assert is_new_stable("0.2.0", "0.3.0b1") != 0
    assert is_new_stable("0.2.0", "0.2.1") == 0
    assert is_new_stable("0.2.0", "0.3.0") == 0

    if len(sys.argv) != 3:
        print(f"{sys.argv[0]} old_stable tag")
        exit(2)

    exit(is_new_stable(sys.argv[1], sys.argv[2]))



import argparse

import demes
import matplotlib.pyplot as plt

import demesdraw


class Command:
    def __init__(self, subparsers, subcommand):
        self.parser = subparsers.add_parser(
            subcommand,
            help=self.__doc__,
            description=self.__doc__,
        )
        self.parser.set_defaults(func=self)
        self.parser.add_argument(
            "--log-time",
            action="store_true",
            default=False,
            help="Use a log-10 scale for the time axis.",
        )
        self.parser.add_argument(
            "--aspect",
            type=float,
            help="Set the aspect ratio (height/width) of the plot.",
        )
        self.parser.add_argument(
            "--scale",
            type=float,
            help="Scale the figure size by the given value.",
        )
        self.parser.add_argument(
            "--title", type=str, help="Set the title of the figure to the given string."
        )
        self.parser.add_argument(
            "input_file",
            metavar="model.yaml",
            type=argparse.FileType(),
            help=(
                "Filename of the model. The special value '-' may be used to "
                "read from stdin. The file may be in any format supported by "
                "the demes library."
            ),
        )
        self.parser.add_argument(
            "output_file",
            metavar="figure.img",
            type=str,
            default=None,
            help=(
                "The filename for the image. The file extension determines "
                "the filetype, and can be any format supported by Matplotlib "
                "(e.g. pdf, svg, png). If no file is specified, "
                "an interactive plot window will be opened."
            ),
            nargs="?",
        )


class TubesCommand(Command):
    """
    Plot a demes-as-tubes schematic of the model and the demes' relationships.
    """

    def __init__(self, subparsers):
        super().__init__(subparsers, "tubes")

        self.parser.add_argument(
            "--scale-bar",
            action="store_true",
            default=False,
            help="Draw a scale bar that indicates population size.",
        )

    def __call__(self, args):
        graph = demes.load(args.input_file)
        fig, ax = demesdraw.utils.get_fig_axes(aspect=args.aspect, scale=args.scale)
        demesdraw.tubes(
            graph,
            ax=ax,
            log_time=args.log_time,
            title=args.title,
            scale_bar=args.scale_bar,
        )
        if args.output_file is not None:
            fig.savefig(args.output_file)
        else:
            # interactive plot
            plt.show()


class SizeHistoryCommand(Command):
    """
    Plot population size as a function of time for each deme in the model.
    """

    def __init__(self, subparsers):
        super().__init__(subparsers, "size_history")

    def __call__(self, args):
        graph = demes.load(args.input_file)
        fig, ax = demesdraw.utils.get_fig_axes(aspect=args.aspect, scale=args.scale)
        demesdraw.size_history(graph, ax=ax, log_time=args.log_time, title=args.title)
        if args.output_file is not None:
            fig.savefig(args.output_file)
        else:
            # interactive plot
            plt.show()


def cli():
    top_parser = argparse.ArgumentParser(
        prog="demesdraw", description="Draw a Demes model."
    )
    top_parser.add_argument(
        "--version", action="version", version=demesdraw.__version__
    )
    subparsers = top_parser.add_subparsers(dest="subcommand")
    TubesCommand(subparsers)
    SizeHistoryCommand(subparsers)
    args = top_parser.parse_args()
    if args.subcommand is None:
        top_parser.print_help()
        exit(1)
    args.func(args)


if __name__ == "__main__":
    cli()



import collections
import itertools
import math

import demes
import pytest
import matplotlib.pyplot as plt
import numpy as np

from demesdraw import utils
import tests


class TestInfStartTime:
    @pytest.mark.filterwarnings("ignore:.*log scale.*:UserWarning:demesdraw.utils")
    @pytest.mark.parametrize("log_scale", [True, False])
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_time_is_reasonable(self, graph, log_scale):
        t = utils._inf_start_time(graph, 0.1, log_scale)
        assert t > 0
        assert not math.isinf(t)

        times = []
        for deme in graph.demes:
            for epoch in deme.epochs:
                times.extend([epoch.start_time, epoch.end_time])
        for migration in graph.migrations:
            times.extend([migration.start_time, migration.end_time])
        for pulse in graph.pulses:
            times.append(pulse.time)
        time_max = max(time for time in times if not math.isinf(time))

        assert t > time_max

    @pytest.mark.parametrize("log_scale", [True, False])
    def test_one_epoch(self, log_scale):
        b = demes.Builder()
        b.add_deme("A", epochs=[dict(start_size=1)])
        graph = b.resolve()
        t = utils._inf_start_time(graph, 0.1, log_scale)
        assert t > 0
        assert not math.isinf(t)


class TestGetFigAxes:
    def teardown_method(self):
        plt.close("all")

    def test_return_values_are_consistent(self):
        fig, ax = utils.get_fig_axes()
        assert fig.get_axes() == [ax]
        assert ax.get_figure() == fig

    @pytest.mark.parametrize("aspect", [0.5, 1, 1.5, 3 / 4, 9 / 16, 10 / 16])
    @pytest.mark.parametrize("scale", [0.5, 1, 1.5])
    def test_aspect_and_scale(self, aspect, scale):
        fig1, _ = utils.get_fig_axes(scale=1, aspect=1)
        width1, height1 = fig1.get_size_inches()
        assert math.isclose(height1, width1)
        fig2, _ = utils.get_fig_axes(scale=scale, aspect=1)
        width2, height2 = fig2.get_size_inches()
        assert math.isclose(height2, width2)
        assert math.isclose(scale, width2 / width1)
        assert math.isclose(scale, height2 / height1)

        fig3, _ = utils.get_fig_axes(scale=1, aspect=aspect)
        width3, height3 = fig3.get_size_inches()
        assert math.isclose(aspect, height3 / width3)
        fig4, _ = utils.get_fig_axes(scale=scale, aspect=aspect)
        width4, height4 = fig4.get_size_inches()
        assert math.isclose(aspect, height4 / width4)
        assert math.isclose(scale, width4 / width3)
        assert math.isclose(scale, height4 / height3)

    def test_tight_layout(self):
        # tight layout is a good default
        fig, _ = utils.get_fig_axes()
        assert fig.get_tight_layout()

    def test_constrained_layout(self):
        # constrained layout should be possible
        fig, _ = utils.get_fig_axes(constrained_layout=True)
        assert not fig.get_tight_layout()
        assert fig.get_constrained_layout()

    def test_multiple_axes(self):
        _, axs = utils.get_fig_axes(nrows=2)
        assert axs.shape == (2,)
        _, axs = utils.get_fig_axes(ncols=2)
        assert axs.shape == (2,)
        _, axs = utils.get_fig_axes(nrows=2, ncols=2)
        assert axs.shape == (2, 2)


class TestSizeMax:
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_size_max(self, graph):
        size_max = utils.size_max(graph)
        for deme in graph.demes:
            for epoch in deme.epochs:
                assert size_max >= epoch.start_size
                assert size_max >= epoch.end_size


class TestSizeMin:
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_size_min(self, graph):
        size_min = utils.size_min(graph)
        for deme in graph.demes:
            for epoch in deme.epochs:
                assert size_min <= epoch.start_size
                assert size_min <= epoch.end_size


class TestLogTimeHeuristic:
    # Don't test the behaviour, just the interface.
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_log_time_heuristic(self, graph):
        log_time = utils.log_time_heuristic(graph)
        assert log_time in [True, False]


class TestLogSizeHeuristic:
    # Don't test the behaviour, just the interface.
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_log_size_heuristic(self, graph):
        log_size = utils.log_size_heuristic(graph)
        assert log_size in [True, False]


class TestInteractionIndices:
    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("n", (1, 2, 3))
    def test_no_interations_isolated_demes(self, n, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            b.add_deme(f"deme{j}")
        graph = b.resolve()
        interactions = utils._interactions_indices(graph, unique=unique)
        assert len(interactions) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_no_interactions_common_ancestor(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"])
        graph = b.resolve()
        interactions = utils._interactions_indices(graph, unique=unique)
        assert len(interactions) == 0

    @pytest.mark.parametrize("n_migrations", (1, 2, 3))
    def test_asymmetric_migrations(self, n_migrations):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"])
        times = np.linspace(0, 100, n_migrations + 1)
        for start_time, end_time in zip(times[1:], times[:-1]):
            b.add_migration(
                source="b",
                dest="c",
                start_time=start_time,
                end_time=end_time,
                rate=1e-5,
            )
        graph = b.resolve()
        interactions = utils._interactions_indices(graph, unique=False)
        assert (
            interactions == [(1, 2)] * n_migrations
            or interactions == [(2, 1)] * n_migrations
        )
        interactions = utils._interactions_indices(graph, unique=True)
        assert interactions == [(1, 2)] or interactions == [(2, 1)]

    @pytest.mark.parametrize("n_migrations", (1, 2, 3))
    def test_symmetric_migrations(self, n_migrations):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"])
        times = np.linspace(0, 100, n_migrations + 1)
        for start_time, end_time in zip(times[1:], times[:-1]):
            b.add_migration(
                demes=["b", "c"], start_time=start_time, end_time=end_time, rate=1e-5
            )
        graph = b.resolve()
        interactions = utils._interactions_indices(graph, unique=False)
        assert len(interactions) == 2 * n_migrations
        counts = collections.Counter(interactions)
        assert len(counts) == 2
        assert counts[(1, 2)] == n_migrations
        assert counts[(2, 1)] == n_migrations
        interactions = utils._interactions_indices(graph, unique=True)
        assert interactions == [(1, 2)] or interactions == [(2, 1)]

    @pytest.mark.parametrize("n_pulses", (1, 2, 3))
    def test_pulses(self, n_pulses):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"])
        for j in range(n_pulses):
            b.add_pulse(
                sources=["b"], dest="c", time=100 * j / n_pulses + 1, proportions=[0.1]
            )
        graph = b.resolve()
        interactions = utils._interactions_indices(graph, unique=False)
        assert (
            interactions == [(1, 2)] * n_pulses or interactions == [(2, 1)] * n_pulses
        )
        interactions = utils._interactions_indices(graph, unique=True)
        assert interactions == [(1, 2)] or interactions == [(2, 1)]


class TestLineCrossings:
    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("n", (1, 2, 3))
    def test_no_crossings_isolated_demes(self, n, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            b.add_deme(f"deme{j}")
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 0
        for x in itertools.permutations(range(n)):
            assert utils._line_crossings(x, candidates) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_no_crossings_common_ancestor(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"])
        b.add_pulse(sources=["b"], dest="c", time=50, proportions=[0.1])
        b.add_migration(demes=["b", "c"], rate=1e-5)
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 0
        for x in itertools.permutations(range(3)):
            assert utils._line_crossings(x, candidates) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestry_crossing(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["b"], start_time=50)
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 1

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 0
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestry_crossing_multiple_ancestors(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["a", "b"], proportions=[0.5, 0.5], start_time=50)
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 2

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 1
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 1

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestry_crossing_treelike(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"], epochs=[dict(end_time=50)])
        b.add_deme("d", ancestors=["c"])
        b.add_deme("e", ancestors=["c"])
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 2

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c", "d", "e"]), candidates) == 0
        assert utils._line_crossings(order(["a", "c", "b", "d", "e"]), candidates) == 2
        assert utils._line_crossings(order(["a", "d", "e", "b", "c"]), candidates) == 2
        assert utils._line_crossings(order(["d", "e", "a", "b", "c"]), candidates) == 2
        assert utils._line_crossings(order(["a", "d", "b", "c", "e"]), candidates) == 1
        assert utils._line_crossings(order(["a", "e", "b", "c", "d"]), candidates) == 1
        assert utils._line_crossings(order(["d", "b", "a", "c", "e"]), candidates) == 1
        assert utils._line_crossings(order(["e", "b", "a", "c", "d"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a", "d", "e"]), candidates) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_pulse_crossing(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_pulse(sources=["b"], dest="c", time=100, proportions=[0.1])
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 1

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 0
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_pulse_crossing_multiple_sources(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_pulse(sources=["a", "b"], dest="c", time=100, proportions=[0.1, 0.1])
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 2

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 1
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 1

    @pytest.mark.parametrize("unique", (True, False))
    def test_asymmetric_migration_crossing(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_migration(source="b", dest="c", rate=1e-5)
        graph = b.resolve()
        candidates = utils._get_line_candidates(graph, unique=unique)
        assert len(candidates) == 1

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 0
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 0

    def test_symmetric_migration_crossing(self):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_migration(demes=["b", "c"], rate=1e-5)
        graph = b.resolve()
        # unique = False
        candidates = utils._get_line_candidates(graph, unique=False)
        assert len(candidates) == 2

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 0
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 2
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 2
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 0

        # unique = True
        candidates = utils._get_line_candidates(graph, unique=True)
        assert len(candidates) == 1

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 0
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 0

    def test_multiple_candidates(self):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["b"], start_time=50)
        b.add_pulse(sources=["a"], dest="c", time=20, proportions=[0.1])
        b.add_migration(source="b", dest="c", rate=1e-5)
        graph = b.resolve()
        # unique = False
        candidates = utils._get_line_candidates(graph, unique=False)
        assert len(candidates) == 3

        def order(deme_names):
            pos = {name: j * 150 for j, name in enumerate(deme_names)}
            return np.array([pos[deme.name] for deme in graph.demes])

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 1
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 2
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 2
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 1

        # Check batch mode gives the same result.
        assert all(
            utils._line_crossings(
                np.stack(
                    [
                        order(["a", "b", "c"]),
                        order(["a", "c", "b"]),
                        order(["b", "a", "c"]),
                        order(["c", "a", "b"]),
                        order(["b", "c", "a"]),
                        order(["c", "b", "a"]),
                    ]
                ),
                candidates,
            )
            == [1, 0, 2, 2, 0, 1]
        )

        # unique = True
        candidates = utils._get_line_candidates(graph, unique=True)
        assert len(candidates) == 2

        assert utils._line_crossings(order(["a", "b", "c"]), candidates) == 1
        assert utils._line_crossings(order(["a", "c", "b"]), candidates) == 0
        assert utils._line_crossings(order(["b", "a", "c"]), candidates) == 1
        assert utils._line_crossings(order(["c", "a", "b"]), candidates) == 1
        assert utils._line_crossings(order(["b", "c", "a"]), candidates) == 0
        assert utils._line_crossings(order(["c", "b", "a"]), candidates) == 1

        # Check batch mode gives the same result.
        assert all(
            utils._line_crossings(
                np.stack(
                    [
                        order(["a", "b", "c"]),
                        order(["a", "c", "b"]),
                        order(["b", "a", "c"]),
                        order(["c", "a", "b"]),
                        order(["b", "c", "a"]),
                        order(["c", "b", "a"]),
                    ]
                ),
                candidates,
            )
            == [1, 0, 1, 1, 0, 1]
        )


class TestMinimalCrossingPositions:
    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("n", (1, 2, 3))
    @pytest.mark.parametrize("sep", (1, math.pi, 12345))
    def test_no_crossings_isolated_demes(self, n, unique, sep):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            b.add_deme(f"deme{j}")
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        for j, deme in enumerate(graph.demes):
            assert math.isclose(positions[deme.name], j * sep)

    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("sep", (1, math.pi, 12345))
    def test_no_crossings_common_ancestor(self, unique, sep):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"])
        b.add_pulse(sources=["b"], dest="c", time=50, proportions=[0.1])
        b.add_migration(demes=["b", "c"], rate=1e-5)
        graph = b.resolve()

        positions = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        for j, deme in enumerate(graph.demes):
            assert math.isclose(positions[deme.name], j * sep)

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestry_crossing(self, unique):
        # Tree-like ancestry.
        # We should never get positions with 'b' in the middle.
        # |\
        # | \
        # |\ \
        # a c b
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["a"], start_time=50)
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )
        assert not (positions["a"] < positions["b"] < positions["c"])
        assert not (positions["c"] < positions["b"] < positions["a"])

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestry_crossing_treelike_uses_default_ordering(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["a"], epochs=[dict(end_time=50)])
        b.add_deme("d", ancestors=["c"])
        b.add_deme("e", ancestors=["c"])
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert order(["a", "b", "c", "d", "e"])

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestry_crossing_multiple_ancestors(self, unique):
        # 'c' has multiple ancestors, so should be in the middle.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["a", "b"], proportions=[0.5, 0.5], start_time=50)
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert order(["a", "c", "b"]) or order(["b", "c", "a"])

    @pytest.mark.parametrize("unique", (True, False))
    def test_pulse_crossing(self, unique):
        # Pulse from 'a' to 'c', so 'b' shouldn't be in the middle.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_pulse(sources=["a"], dest="c", time=100, proportions=[0.1])
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert not (order(["a", "b", "c"]) or order(["c", "b", "a"]))

    @pytest.mark.parametrize("unique", (True, False))
    def test_pulse_crossing_multiple_sources(self, unique):
        # Pulses from 'a' and 'b' into 'c', so 'c' should be in the middle.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_pulse(sources=["a", "b"], dest="c", time=100, proportions=[0.1, 0.1])
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert order(["a", "c", "b"]) or order(["b", "c", "a"])

    @pytest.mark.parametrize("unique", (True, False))
    def test_asymmetric_migration_crossing(self, unique):
        # Migration from 'a' to 'c', so 'b' shouldn't be in the middle.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_migration(source="a", dest="c", rate=1e-5)
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert not (order(["a", "b", "c"]) or order(["c", "b", "a"]))

    @pytest.mark.parametrize("unique", (True, False))
    def test_symmetric_migration_crossing(self, unique):
        # Migration from 'a' to 'c', so 'b' shouldn't be in the middle.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b")
        b.add_deme("c")
        b.add_migration(demes=["a", "c"], rate=1e-5)
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert not (order(["a", "b", "c"]) or order(["c", "b", "a"]))

    @pytest.mark.parametrize("unique", (True, False))
    def test_ancestor_pulse_migration(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["b"], start_time=50)
        b.add_pulse(sources=["a"], dest="c", time=20, proportions=[0.1])
        b.add_migration(source="b", dest="c", rate=1e-5)
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        assert order(["a", "c", "b"]) or order(["b", "c", "a"])

    def test_unique_interactions(self):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["b"], start_time=50)
        b.add_pulse(sources=["a"], dest="c", time=40, proportions=[0.1])
        b.add_pulse(sources=["a"], dest="c", time=30, proportions=[0.1])
        b.add_pulse(sources=["a"], dest="b", time=40, proportions=[0.1])
        b.add_pulse(sources=["a"], dest="b", time=30, proportions=[0.1])
        graph = b.resolve()

        def order(deme_names):
            b = True
            for j in range(len(deme_names) - 1):
                b = b and (positions[deme_names[j]] < positions[deme_names[j + 1]])
            return b

        # 'a' shouldn't be in the middle.
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=True
        )
        assert not (order(["b", "a", "c"]) or order(["c", "a", "b"]))

        # 'a' should be in the middle.
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=False
        )
        assert order(["b", "a", "c"]) or order(["c", "a", "b"]), positions

    @pytest.mark.parametrize("unique", (True, False))
    def test_many_demes(self, unique):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(20):
            b.add_deme(f"deme{j}")
        b.add_pulse(sources=["deme0"], dest="deme2", time=10, proportions=[1.0])
        graph = b.resolve()
        positions = utils.minimal_crossing_positions(
            graph, sep=1, unique_interactions=unique
        )

        # deme0 and deme2 should be adjacent.
        assert not positions["deme0"] < positions["deme1"] < positions["deme2"]
        for j in range(3, 20):
            assert not positions["deme0"] < positions[f"deme{j}"] < positions["deme2"]

    @pytest.mark.parametrize("graph", tests.example_graphs())
    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("sep", (1, math.pi, 12345))
    def test_separation_distance(self, graph, unique, sep):
        positions = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        for d1 in graph.demes:
            for d2 in graph.demes:
                if d1 is d2:
                    continue
                assert abs(positions[d1.name] - positions[d2.name]) >= sep


class TestCoexistenceIndices:
    @pytest.mark.parametrize("n", (1, 2, 3))
    def test_n_immortal_demes(self, n):
        # All pairs of demes coexist.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            b.add_deme(f"deme{j}")
        graph = b.resolve()
        idx = utils._coexistence_indices(graph)
        assert set(idx) == set(itertools.combinations(range(n), 2))

    @pytest.mark.parametrize("n", (1, 2, 3))
    def test_succesive_descendents(self, n):
        # Only one deme exists at any given time.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            ancestors = []
            if j != 0:
                ancestors = [f"deme{j-1}"]
            b.add_deme(f"deme{j}", ancestors=ancestors, epochs=[dict(end_time=100 - j)])
        graph = b.resolve()
        idx = utils._coexistence_indices(graph)
        assert len(idx) == 0

    @pytest.mark.parametrize("unique", (True, False))
    def test_tree_like1(self, unique):
        # Tree-like ancestry.
        # |
        # |\
        # | |\
        # a b c
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a")
        b.add_deme("b", ancestors=["a"], start_time=100)
        b.add_deme("c", ancestors=["b"], start_time=50)
        graph = b.resolve()
        idx = utils._coexistence_indices(graph)
        assert len(idx) == 3
        assert (0, 1) in idx
        assert (0, 2) in idx
        assert (1, 2) in idx

    @pytest.mark.parametrize("unique", (True, False))
    def test_tree_like2(self, unique):
        # Tree-like ancestry.
        # |
        # a
        # |\
        # | |\
        # b c d
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        b.add_deme("a", epochs=[dict(end_time=100)])
        b.add_deme("b", ancestors=["a"])
        b.add_deme("c", ancestors=["b"], start_time=50)
        b.add_deme("d", ancestors=["c"], start_time=10)
        graph = b.resolve()
        idx = utils._coexistence_indices(graph)
        assert len(idx) == 3
        assert (1, 2) in idx
        assert (1, 3) in idx
        assert (2, 3) in idx


class TestOptimisePositions:
    @pytest.mark.parametrize("graph", tests.example_graphs())
    @pytest.mark.parametrize("unique", (True, False))
    def test_ordering_doesnt_change(self, graph, unique):
        sep = 1
        positions1 = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        positions2 = utils.optimise_positions(
            graph, positions=positions1, sep=sep, unique_interactions=unique
        )
        for j, k in utils._coexistence_indices(graph):
            if positions1[graph.demes[j].name] < positions1[graph.demes[j].name]:
                assert positions2[graph.demes[j].name] < positions2[graph.demes[j].name]
            elif positions1[graph.demes[j].name] > positions1[graph.demes[j].name]:
                assert positions2[graph.demes[j].name] > positions2[graph.demes[j].name]

    @pytest.mark.parametrize("graph", tests.example_graphs())
    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("sep", (1, math.pi, 12345))
    def test_separation_distance(self, graph, unique, sep):
        positions1 = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        positions2 = utils.optimise_positions(
            graph, positions=positions1, sep=sep, unique_interactions=unique
        )
        epsilon = 1e-3  # Small amount of wiggle room for numerical error.
        for j, k in utils._coexistence_indices(graph):
            assert (
                abs(positions2[graph.demes[j].name] - positions2[graph.demes[k].name])
                >= sep - epsilon
            )

    @pytest.mark.parametrize("n", range(5, 10, 15))
    @pytest.mark.parametrize("unique", (True, False))
    @pytest.mark.parametrize("migrations", (True, False))
    def test_island_model(self, n, unique, migrations):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            b.add_deme(f"d{j}")
            if migrations and j > 0:
                b.add_migration(demes=[f"d{j - 1}", f"d{j}"], rate=1e-5)
        graph = b.resolve()

        sep = utils.separation_heuristic(graph)
        positions1 = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        positions2 = utils.optimise_positions(
            graph, positions=positions1, sep=sep, unique_interactions=unique
        )
        epsilon = 1e-3  # Small amount of wiggle room for numerical error.
        for j, k in utils._coexistence_indices(graph):
            assert (
                abs(positions2[graph.demes[j].name] - positions2[graph.demes[k].name])
                >= sep - epsilon
            )

    @pytest.mark.parametrize("graph", tests.example_graphs())
    @pytest.mark.parametrize("unique", (True, False))
    def test_optimise_gradients(self, graph, unique):
        # Check that the Jacobian and Hessian of the objective function
        # are calculated correctly by comparing them against a numerical
        # approximation.
        # The mpmath package is used to obtain good approximations.
        # This test isn't comprehensive---it only compares the gradients
        # at the starting position x0.
        sep = utils.separation_heuristic(graph)
        positions = utils.minimal_crossing_positions(
            graph, sep=sep, unique_interactions=unique
        )
        successors = utils._successors_indices(graph)
        interactions = utils._interactions_indices(graph, unique=unique)

        x0 = np.array([positions[deme.name] for deme in graph.demes])
        # Place the first deme at position 0.
        x0 -= x0[0]

        obj = utils._PositionsObjective(successors, interactions)

        def f(*x):
            # transformed for mpmath.diff
            return obj.f(x)

        np.testing.assert_almost_equal(obj.g(x0), jacobian(f, x0))
        np.testing.assert_almost_equal(obj.h(x0), hessian(f, x0))


def jacobian(f, x):
    """
    Jacobian of function f, evaluated at x.
    """
    from mpmath import mp

    eye = np.eye(len(x), dtype=int)
    jac = [mp.diff(f, x, n=row) for row in eye]
    return jac


def hessian(f, x):
    """
    Hessian (matrix of second partial derivatives) of function f, evaluated at x.
    """
    from mpmath import mp

    n = len(x)
    H = np.empty((n, n), dtype=float)
    for i in range(n):
        for j in range(n):
            # multi-index for second derivative wrt i and j
            order = tuple(
                ((1 if k == i else 0) + (1 if k == j else 0)) for k in range(n)
            )
            H[i, j] = mp.diff(f, x, n=order)
    return H



import warnings
import demes
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from typing import cast
import numpy as np
import matplotlib

import demesdraw
from demesdraw import utils
from tests import example_files


with PdfPages("spacing.pdf") as pdf:
    for _n in range(30):
        n = _n + 1
        print(n)
        b = demes.Builder(defaults=dict(epoch=dict(start_size=100)))
        for j in range(n):
            b.add_deme(f"d{j}")
            if j > 0:
                b.add_migration(demes=[f"d{j - 1}", f"d{j}"], rate=1e-5)
        graph = b.resolve()

        colours = None
        if n > 20:
            colours = "black"
        ax = demesdraw.tubes(graph, colours=colours)
        pdf.savefig(ax.figure)
        # Ensure mypy sees a concrete Figure (SubFigure may be present at runtime).
        plt.close(cast(matplotlib.figure.Figure, ax.figure))

with PdfPages("examples.pdf") as pdf:
    for filename in sorted(example_files()):
        title = filename.name
        print(title)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Multiple pulses", UserWarning, "demes")
            graph = demes.load(filename)

        fig, ax = utils.get_fig_axes(scale=1.5, nrows=2, constrained_layout=True)
        # ax may be a single Axes or an array/sequence of Axes (depending on kwargs).
        if isinstance(ax, (list, tuple, np.ndarray)):
            ax1, ax2 = ax
        else:
            ax1 = ax2 = ax
        # Help mypy by casting to a concrete Axes type.
        ax1 = cast(matplotlib.axes.Axes, ax1)
        ax2 = cast(matplotlib.axes.Axes, ax2)

        log_size = utils.log_size_heuristic(graph)
        log_time = utils.log_time_heuristic(graph)

        demesdraw.size_history(
            graph,
            ax=ax1,
            invert_x=True,
            log_time=log_time,
            log_size=log_size,
            title=title,
        )
        demesdraw.tubes(
            graph,
            ax=ax2,
            log_time=log_time,
            # fill=False,
            # colours="black",
            scale_bar=True,
        )

        pdf.savefig(fig)
        plt.close(fig)



import pytest

from demesdraw import *


def test_public_symbols():
    size_history
    tubes


def test_nonpublic_symbols():
    with pytest.raises(NameError):
        utils


from demesdraw.utils import *  # noqa: E402


def test_utils_public_symbols():
    get_fig_axes
    size_max
    size_min
    log_size_heuristic
    log_time_heuristic
    separation_heuristic


def test_utils_nonpublic_symbols():
    with pytest.raises(NameError):
        _get_colours


def test_dir():
    import demesdraw

    dir_demesdraw = dir(demesdraw)
    assert "size_history" in dir_demesdraw
    assert "tubes" in dir_demesdraw

    assert "utils" not in dir_demesdraw


def test_utils_dir():
    import demesdraw.utils

    dir_ddu = dir(demesdraw.utils)
    assert "get_fig_axes" in dir_ddu
    assert "size_max" in dir_ddu
    assert "size_min" in dir_ddu
    assert "log_size_heuristic" in dir_ddu
    assert "log_time_heuristic" in dir_ddu

    assert "_get_colours" not in dir_ddu
    assert "size_history" not in dir_ddu
    assert "tubes" not in dir_ddu



import math

import demes
import pytest
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.testing.compare import compare_images
from matplotlib.backend_bases import MouseEvent
import numpy as np

import demesdraw
import tests


def images_equal(file1, file2):
    # Matplotlib's compare_images() function has a weird return value.
    # This wrapper just makes test assertions clearer.
    cmp = compare_images(file1, file2, tol=0)
    return cmp is None


class TestSizeHistory:
    def check_size_history(self, graph, **kwargs):
        ax = demesdraw.size_history(graph, **kwargs)
        assert isinstance(ax, matplotlib.axes.Axes)
        plt.close(ax.figure)

    @pytest.mark.filterwarnings("ignore:.*log scale.*:UserWarning:demesdraw.utils")
    @pytest.mark.parametrize("log_size", [True, False])
    @pytest.mark.parametrize("log_time", [True, False])
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_log_params(self, graph, log_time, log_size):
        self.check_size_history(graph, log_time=log_time, log_size=log_size)

    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_annotate_epochs_invert_x(self, graph):
        self.check_size_history(graph, invert_x=True, annotate_epochs=True)


class TestTubes:
    def check_tubes(self, graph, seed=1234, **kwargs):
        ax = demesdraw.tubes(graph, seed=seed, **kwargs)
        assert isinstance(ax, matplotlib.axes.Axes)
        plt.close(ax.figure)

    @pytest.mark.filterwarnings("ignore:.*log scale.*:UserWarning:demesdraw.utils")
    @pytest.mark.parametrize("log_time", [True, False])
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_log_params(self, graph, log_time):
        self.check_tubes(graph, log_time=log_time)

    @pytest.mark.parametrize(
        "labels", ["xticks", "legend", "mid", "xticks-legend", "xticks-mid"]
    )
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_labels_params(self, graph, labels):
        self.check_tubes(graph, labels=labels)

    @pytest.mark.parametrize("scale_bar", [True, False])
    @pytest.mark.parametrize("graph", tests.example_graphs())
    def test_scale_bar(self, graph, scale_bar):
        self.check_tubes(graph, scale_bar=scale_bar)

    @pytest.mark.parametrize(
        "colours", [None, "black", dict(), dict(A="red", B="blue", C="pink")]
    )
    def test_colours_params(self, colours):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A", epochs=[dict(end_time=100)])
        b.add_deme("B", ancestors=["A"])
        b.add_deme("C", ancestors=["A"])
        graph = b.resolve()
        self.check_tubes(graph, colours=colours)

        for j in range(10):
            b.add_deme(f"deme{j}", ancestors=["A"])
        graph = b.resolve()
        # 13 demes. Also fine.
        self.check_tubes(graph, colours=colours)

    def test_bad_colours_params(self):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A", epochs=[dict(end_time=100)])
        b.add_deme("B", ancestors=["A"])
        b.add_deme("C", ancestors=["A"])
        graph = b.resolve()
        for colour in [object(), math.inf, "thisisnotarealcolour"]:
            with pytest.raises(ValueError, match="not.*a matplotlib colour"):
                self.check_tubes(graph, colours=colour)
        with pytest.raises(ValueError, match="deme.*not found in the graph"):
            self.check_tubes(graph, colours=dict(X="black"))

        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        for j in range(25):
            b.add_deme(f"deme{j}")
        b.add_migration(demes=[f"deme{j}" for j in range(25)], rate=1e-5)
        graph = b.resolve()
        with pytest.raises(ValueError, match="colours must be specified"):
            self.check_tubes(graph)

    @pytest.mark.parametrize("log_time", [True, False])
    def test_max_time(self, log_time):
        def check_max_time(graph):
            for max_time in [50, 5000, 1e7]:
                ax = demesdraw.tubes(graph, log_time=log_time, max_time=max_time)
                ylim = ax.get_ylim()
                assert math.isclose(ylim[1], max_time)
                plt.close(ax.figure)

        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A")
        b.add_deme("B")
        graph = b.resolve()
        check_max_time(graph)

        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A", epochs=[dict(end_time=100)])
        b.add_deme("B", ancestors=["A"])
        b.add_deme("C", ancestors=["A"])
        graph = b.resolve()
        check_max_time(graph)

    def test_format_coord(self):
        # The Axes object's format_coord() function gives the text for the
        # status bar in interactive figures. We want this to indicate the
        # time corresponding to the mouse cursor location.
        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A")
        graph = b.resolve()
        ax = demesdraw.tubes(graph)
        w, h = ax.figure.canvas.get_width_height()
        status_text = ax.format_coord(w / 2, h / 2)
        assert "x" not in status_text
        assert "y" not in status_text
        assert "time" in status_text

    @pytest.mark.parametrize("log_time", [True, False])
    @pytest.mark.usefixtures("tmp_path")
    def test_mouseover_deme(self, log_time, tmp_path):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A")
        graph = b.resolve()

        ax = demesdraw.tubes(graph, log_time=log_time)
        ax.figure.savefig(tmp_path / "fig1.png")

        # Simulate mouseover in the middle of the figure.
        # An annotation should appear, showing deme info.
        w, h = ax.figure.canvas.get_width_height()
        event = MouseEvent("motion_notify_event", ax.figure.canvas, w / 2, h / 2)
        ax.figure.canvas.callbacks.process("motion_notify_event", event)
        ax.figure.savefig(tmp_path / "fig2.png")

        assert not images_equal(tmp_path / "fig1.png", tmp_path / "fig2.png")

        # Simulate mouseover at the edge of the figure.
        # The annotation should disappear.
        event = MouseEvent("motion_notify_event", ax.figure.canvas, 1, 1)
        ax.figure.canvas.callbacks.process("motion_notify_event", event)
        ax.figure.savefig(tmp_path / "fig3.png")

        assert images_equal(tmp_path / "fig1.png", tmp_path / "fig3.png")

    @pytest.mark.parametrize("log_time", [True, False])
    @pytest.mark.usefixtures("tmp_path")
    def test_mouseover_pulse(self, log_time, tmp_path):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A")
        b.add_deme("B")
        b.add_pulse(sources=["A"], dest="B", proportions=[0.1], time=100)
        graph = b.resolve()

        ax = demesdraw.tubes(graph, log_time=log_time)
        ax.figure.savefig(tmp_path / "fig1.png")

        # Simulate moving the mouse over the pulse lines.
        # An annotation should appear, showing pulse info.
        num_pulse_lines = 0
        for line in ax.lines:
            if not (
                hasattr(line, "_demesdraw_data") and "pulse" in line._demesdraw_data
            ):
                continue
            xs, ys = line.get_xdata(), line.get_ydata()
            x, y = ax.transData.transform((np.mean(xs), ys[0]))
            event = MouseEvent("motion_notify_event", ax.figure.canvas, x, y)
            ax.figure.canvas.callbacks.process("motion_notify_event", event)
            ax.figure.savefig(tmp_path / "fig2.png")

            assert not images_equal(tmp_path / "fig1.png", tmp_path / "fig2.png")
            num_pulse_lines += 1
        assert num_pulse_lines > 0

        # Simulate mouseover at the edge of the figure.
        # The annotation should disappear.
        event = MouseEvent("motion_notify_event", ax.figure.canvas, 1, 1)
        ax.figure.canvas.callbacks.process("motion_notify_event", event)
        ax.figure.savefig(tmp_path / "fig3.png")

        assert images_equal(tmp_path / "fig1.png", tmp_path / "fig3.png")

    @pytest.mark.parametrize("log_time", [True, False])
    @pytest.mark.usefixtures("tmp_path")
    def test_mouseover_migration(self, log_time, tmp_path):
        b = demes.Builder(defaults=dict(epoch=dict(start_size=1000)))
        b.add_deme("A")
        b.add_deme("B")
        b.add_migration(demes=["A", "B"], rate=1e-5)
        graph = b.resolve()

        ax = demesdraw.tubes(graph, log_time=log_time)
        ax.figure.savefig(tmp_path / "fig1.png")

        # Simulate moving the mouse over the migration lines.
        # An annotation should appear, showing migration info.
        num_migration_lines = 0
        for line in ax.lines:
            if not (
                hasattr(line, "_demesdraw_data") and "migration" in line._demesdraw_data
            ):
                continue
            xs, ys = line.get_xdata(), line.get_ydata()
            x, y = ax.transData.transform((np.mean(xs), ys[0]))
            event = MouseEvent("motion_notify_event", ax.figure.canvas, x, y)
            ax.figure.canvas.callbacks.process("motion_notify_event", event)
            ax.figure.savefig(tmp_path / "fig2.png")

            assert not images_equal(tmp_path / "fig1.png", tmp_path / "fig2.png")
            num_migration_lines += 1
        assert num_migration_lines > 0

        # Simulate mouseover at the edge of the figure.
        # The annotation should disappear.
        event = MouseEvent("motion_notify_event", ax.figure.canvas, 1, 1)
        ax.figure.canvas.callbacks.process("motion_notify_event", event)
        ax.figure.savefig(tmp_path / "fig3.png")

        assert images_equal(tmp_path / "fig1.png", tmp_path / "fig3.png")



"""
Add version string to the navbar and footer.
"""

import demesdraw


def inject_version(app, config):
    v = demesdraw.__version__
    if v != "undefined":
        v_short = v.split("+")[0]
        config.html_theme_options["extra_navbar"] = f"demesdraw {v_short}"
        config.html_theme_options["extra_footer"] = f"demesdraw {v}"


def setup(app):
    app.connect("config-inited", inject_version)



import pathlib
import subprocess
import tempfile

import pytest

import demesdraw
import tests


class TestToplevel:
    def test_help(self):
        out1 = subprocess.run(
            "python -m demesdraw -h".split(), check=True, stdout=subprocess.PIPE
        )
        out2 = subprocess.run(
            "python -m demesdraw --help".split(), check=True, stdout=subprocess.PIPE
        )
        assert out1.stdout == out2.stdout

    def test_no_arguments_produces_help_output(self):
        # If no params are given, the output is the same as --help.
        # But the returncode should be non-zero.
        out1 = subprocess.run("python -m demesdraw -h".split(), stdout=subprocess.PIPE)
        out2 = subprocess.run("python -m demesdraw".split(), stdout=subprocess.PIPE)
        assert out1.stdout == out2.stdout
        assert out2.returncode != 0

    def test_version(self):
        out = subprocess.run(
            "python -m demesdraw --version".split(),
            stdout=subprocess.PIPE,
            encoding="utf8",
        )
        assert out.stdout.strip() == demesdraw.__version__


class TestSubCommand:
    @pytest.mark.parametrize("subcommand", ["tubes", "size_history"])
    def test_help(self, subcommand):
        out1 = subprocess.run(
            f"python -m demesdraw {subcommand} -h".split(),
            check=True,
            stdout=subprocess.PIPE,
        )
        out2 = subprocess.run(
            f"python -m demesdraw {subcommand} --help".split(),
            check=True,
            stdout=subprocess.PIPE,
        )
        assert out1.stdout == out2.stdout

    @pytest.mark.parametrize(
        "params",
        [
            "",
            "--log-time",
            "--aspect 1",
            "--scale 1.2",
            "--aspect 1 --scale 1.2 --log-time",
        ],
    )
    @pytest.mark.parametrize("subcommand", ["tubes", "size_history"])
    def test_output_file_is_created(self, subcommand, params):
        input_file = tests.example_files()[0]
        with tempfile.TemporaryDirectory() as tmpdir:
            output_file = pathlib.Path(tmpdir) / "plot.pdf"
            subprocess.run(
                f"python -m demesdraw {subcommand} {params} "
                f"{input_file} {output_file}".split(),
                check=True,
            )
            assert output_file.exists()
            assert output_file.stat().st_size > 0

    def test_tubes_scale_bar(self):
        self.test_output_file_is_created("tubes", "--scale-bar")

    @pytest.mark.parametrize("subcommand", ["tubes", "size_history"])
    def test_input_from_stdin(self, subcommand):
        input_file = tests.example_files()[0]
        with tempfile.TemporaryDirectory() as tmpdir:
            output_file = pathlib.Path(tmpdir) / "plot.pdf"
            with open(input_file) as f:
                subprocess.run(
                    f"python -m demesdraw {subcommand} - {output_file}".split(),
                    check=True,
                    stdin=f,
                )
            assert output_file.exists()
            assert output_file.stat().st_size > 0



__version__ = "undefined"
try:
    from . import _version

    __version__ = _version.version
except ImportError:
    pass

from .size_history import size_history
from .tubes import tubes

__all__ = ["size_history", "tubes"]


# Override the symbols that are returned when calling dir(<module-name>).
def __dir__():
    return sorted(__all__)



from __future__ import annotations
from typing import Dict, Mapping, List, Tuple

import demes
import numpy as np
import matplotlib
import matplotlib.patheffects

from . import utils


class Tube:
    """
    A deme represented as a tube. The tube has a length along the time
    dimension and a width equal to the deme's population size, which may
    change over time.

    :ivar list[float] time: Coordinates along the time dimension.
    :ivar list[float] size1: Coordinates along the non-time dimension,
        corresponding to the first side of the tube.
    :ivar list[float] size2: Coordinates along the non-time dimension,
        corresponding to the second side of the tube.
    """

    def __init__(
        self,
        deme: demes.Deme,
        mid: float,
        inf_start_time: float,
        log_time: bool = False,
    ):
        """
        :param demes.Deme deme: The deme for which to calculate coordinates.
        :param float mid: The mid point of the deme along the non-time dimension.
        :param float inf_start_time: The value along the time dimension which
            is used instead of infinity (for epochs with infinite start times).
        :param bool log_time: The time axis uses a log-10 scale.
        """
        self.deme = deme
        self.mid = mid
        self._coords(deme, mid, inf_start_time, log_time)

    def _coords(
        self,
        deme: demes.Deme,
        mid: float,
        inf_start_time: float,
        log_time: bool,
        num_points: int = 100,
    ) -> None:
        """Calculate tube coordinates."""
        time: List[float] = []
        size1: List[float] = []
        size2: List[float] = []
        for k, epoch in enumerate(deme.epochs):
            start_time = epoch.start_time
            if np.isinf(start_time):
                start_time = inf_start_time
            end_time = epoch.end_time

            if epoch.size_function == "constant":
                t = np.array([start_time, end_time])
                N1 = np.array([mid - epoch.start_size / 2] * 2)
                N2 = np.array([mid + epoch.end_size / 2] * 2)
            elif epoch.size_function == "exponential":
                if log_time:
                    t = np.exp(
                        np.linspace(
                            np.log(start_time), np.log(max(1, end_time)), num=num_points
                        )
                    )
                else:
                    t = np.linspace(start_time, end_time, num=num_points)
                dt = (start_time - t) / (start_time - end_time)
                r = np.log(epoch.end_size / epoch.start_size)
                N = epoch.start_size * np.exp(r * dt)
                N1 = mid - N / 2
                N2 = mid + N / 2
            elif epoch.size_function == "linear":
                if log_time:
                    t = np.exp(
                        np.linspace(
                            np.log(start_time), np.log(max(1, end_time)), num=num_points
                        )
                    )
                else:
                    t = np.linspace(start_time, end_time, num=num_points)
                dt = (start_time - t) / (start_time - end_time)
                N = epoch.start_size + (epoch.end_size - epoch.start_size) * dt
                N1 = mid - N / 2
                N2 = mid + N / 2
            else:
                raise ValueError(
                    f"Don't know how to draw epoch {k} with "
                    f'"{epoch.size_function}" size_function.'
                )

            time.extend(t)
            size1.extend(N1)
            size2.extend(N2)

        self.time = time
        self.size1 = size1
        self.size2 = size2

    def sizes_at(self, time: float) -> Tuple[float, float]:
        """Return the size coordinates of the tube at the given time."""
        N = self.deme.size_at(time)
        N1 = self.mid - N / 2
        N2 = self.mid + N / 2
        return N1, N2

    def to_path(self) -> matplotlib.path.Path:
        """Return a matplotlib.path.Path for the tube outline."""
        # Compared with separately drawing each side of the tube,
        # drawing a path that includes both sides of the tube means:
        #  * vector formats use one "stroke" rather than two; and
        #  * matplotlib path effects apply once, avoiding artifacts for
        #    very narrow tubes, where the path effect from the side
        #    drawn second is on top of the line drawn for the first side.
        N = list(self.size1) + list(self.size2)
        t = list(self.time) * 2
        vertices = list(zip(N, t))
        codes = (
            [matplotlib.path.Path.MOVETO]
            + [matplotlib.path.Path.LINETO] * (len(self.time) - 1)
        ) * 2
        return matplotlib.path.Path(vertices, codes)


def find_positions(graph: demes.Graph) -> Dict[str, float]:
    """
    Find optimal deme positions along a single dimension by:

      - ordering the demes to minimise the number of lines crossing demes,
      - minimising the distance from each parent deme to the mean position of
        its children,
      - minimising the distance between interacting demes (where interactions
        are either migrations or pulses).

    In addition, the solution is constrained so that contemporary demes
    have a minimum separation distance.

    :param graph:
        The graph for which positions should be obtained.
    :return:
        A dictionary mapping deme names to positions.
    """
    sep = utils.separation_heuristic(graph)
    positions = utils.minimal_crossing_positions(
        graph, sep=sep, unique_interactions=False
    )
    return utils.optimise_positions(
        graph, positions=positions, sep=sep, unique_interactions=False
    )


def tubes(
    graph: demes.Graph,
    ax: matplotlib.axes.Axes | None = None,
    colours: utils.ColourOrColourMapping | None = None,
    log_time: bool = False,
    title: str | None = None,
    inf_ratio: float = 0.2,
    positions: Mapping[str, float] | None = None,
    num_lines_per_migration: int = 10,
    seed: int | None = None,
    max_time: float | None = None,
    labels: str = "xticks-mid",
    fill: bool = True,
    scale_bar: bool = False,
) -> matplotlib.axes.Axes:
    """
    Plot a demes-as-tubes schematic of the graph and the demes' relationships.

    Each deme is depicted as a tube, where the tube’s width is
    proportional to the deme’s size at any given time.
    Horizontal lines with arrows indicate either:

     * an ancestor/descendant relation (thick solid lines, open arrow heads),
     * an admixture pulse (dashed lines, closed arrow heads),
     * or a period of continuous migration (thin solid lines, closed arrow heads).

    Lines are drawn in the colour of the ancestor or source deme, and arrows
    point from ancestor to descendant or from source to dest.
    For each period of continuous migration, multiple thin lines are drawn.
    The time for each migration line is sampled uniformly at
    random from the migration's time interval (or log-uniformly for a
    log-scaled time axis). Symmetric migrations have lines in both directions.

    If ``positions`` are not specified, the positions will be chosen
    automatically such that line crossings are minimised and related demes
    are close together. If the automatically chosen positions are unexpected,
    please open an issue at https://github.com/grahamgower/demesdraw/issues/.

    :param demes.Graph graph:
        The demes graph to plot.
    :param ax:
        The matplotlib axes onto which the figure
        will be drawn. If None, an empty axes will be created for the figure.
    :type ax: Optional[matplotlib.axes.Axes]
    :param colours:
        A mapping from deme name to matplotlib colour. Alternately,
        ``colours`` may be a named colour that will be used for all demes.
    :type colours: Optional[dict or str]
    :param log_time:
        If True, use a log-10 scale for the time axis.
        If False (*default*), a linear scale will be used.
    :param title:
        The title of the figure.
    :param inf_ratio:
        The proportion of the time axis that will be
        used for the time interval which stretches towards infinity.
    :param positions:
        A dictionary mapping deme names to horizontal coordinates
        (the mid point of the deme's tube).
        Note that the width of a deme is the deme's (max) size,
        so the positions should allow sufficient space to avoid overlapping.
    :param num_lines_per_migration:
        The number of lines to draw per migration. For symmetric migrations,
        this number of lines will be drawn in each direction.
    :param seed:
        Seed for the random number generator. The generator is
        used to sample times for migration lines.
    :param max_time:
        The maximum time value shown in the figure.
        If demographic events (e.g. size changes, migrations, common ancestor
        events) occur before this time, those events will not be visible.
        If no demographic events occur before this time, the root demes will be
        drawn so they extend to the given time.
    :param labels:
        A string indicating where the deme names should be drawn, or ``None``.
        The options are:

         * "xticks": for extant demes, labels are written under the x-axis
           as matplotlib xticklabels. Labels are not written for extinct demes.
         * "legend": labels for colour patches are written in a legend.
         * "mid": labels are written in the middle of each deme's tube.
         * "xticks-legend": for extant demes, labels are written under the
           x-axis as matplotlib xticklabels. For extinct demes, labels for
           colour patches are written in a legend.
         * "xticks-mid" (*default*): for extant demes, labels are written
           under the x-axis as matplotlib xticklabels. For extinct demes,
           labels are written in the middle of each deme's tube.
         * ``None``: no labels are written.

    :type labels: str or None
    :param fill:
        If True, the inside of the tubes will be painted.
        If False, only the outline of the tubes will be drawn.
    :param scale_bar:
        If True, draw a scale bar that indicates population size.
        If False (*default*), no scale bar will be shown.
    :return:
        The matplotlib axes onto which the figure was drawn.
    """
    if labels not in ("xticks", "legend", "mid", "xticks-legend", "xticks-mid", None):
        raise ValueError(f"Unexpected value for labels: '{labels}'")

    if ax is None:
        _, ax = utils.get_fig_axes()

    if log_time:
        ax.set_yscale("log", base=10)

    colours = utils._get_colours(graph, colours)

    rng = np.random.default_rng(seed)
    inf_start_time = max_time
    if inf_start_time is None:
        inf_start_time = utils._inf_start_time(graph, inf_ratio, log_time)

    if positions is None:
        positions = find_positions(graph)

    tubes = {}
    ancestry_arrows = []

    for j, deme in enumerate(graph.demes):
        colour = colours[deme.name]
        plot_kwargs = dict(
            color=colour,
            # solid_capstyle="butt",
            # capstyle="butt",
            zorder=1,
            path_effects=[
                matplotlib.patheffects.withStroke(linewidth=3, foreground="white")
            ],
        )

        mid = positions[deme.name]

        tube = Tube(deme, mid, inf_start_time, log_time=log_time)
        tubes[deme.name] = tube

        path_patch = matplotlib.patches.PathPatch(
            tube.to_path(), capstyle="butt", fill=False, **plot_kwargs
        )
        ax.add_patch(path_patch)
        if fill:
            poly = ax.fill_betweenx(
                tube.time,
                tube.size1,
                tube.size2,
                facecolor=colour,
                edgecolor="none",
                alpha=0.5,
                zorder=1,
            )
            poly._demesdraw_data = dict(deme=deme)  # type: ignore[attr-defined]

        # Indicate ancestry from ancestor demes.
        tube_frac = np.linspace(
            tube.size1[0], tube.size2[0], max(2, len(deme.ancestors))
        )
        left = 0
        right = 0
        arrow_kwargs = dict(
            arrowstyle="-|>",
            mutation_scale=10,
            alpha=1,
            facecolor=(1, 1, 1, 0),
            path_effects=[
                matplotlib.patheffects.withStroke(linewidth=3, foreground="white")
            ],
            zorder=2,
        )
        for ancestor_id, proportion in zip(deme.ancestors, deme.proportions):
            anc_size1, anc_size2 = tubes[ancestor_id].sizes_at(deme.start_time)
            if anc_size2 < tube.size1[0]:
                # Ancestor is to the left.
                x_pos = [anc_size2, tube_frac[left]]
                ancestry_arrows.append(
                    (
                        tube.time[0],
                        min(x_pos),
                        max(x_pos),
                        colours[ancestor_id],
                    )
                )
                left += 1
            elif tube.size2[0] < anc_size1:
                # Ancestor is to the right.
                x_pos = [anc_size1, tube_frac[len(tube_frac) - right - 1]]
                ancestry_arrows.append(
                    (
                        tube.time[0],
                        max(x_pos),
                        min(x_pos),
                        colours[ancestor_id],
                    )
                )
                right += 1

    slots: Dict[int, int] = {}
    for j, (time, x1, x2, colour) in enumerate(ancestry_arrows):
        taken = set()
        for k, slot in slots.items():
            k_time, k_x1, k_x2, _ = ancestry_arrows[k]
            if np.isclose(time, k_time) and not (
                min(x1, x2) > max(k_x1, k_x2) or min(k_x1, k_x2) > max(x1, x2)
            ):
                taken.add(slot)
        slot = 0
        i = 0
        while slot in taken:
            i += 1
            odd = i % 2
            delta = i // 2
            slot = -odd * delta + (1 - odd) * delta
        slots[j] = slot
        radius = slot * 0.15
        arr = matplotlib.patches.FancyArrowPatch(
            (x1, time),
            (x2, time),
            connectionstyle=f"arc3,rad={radius}",
            edgecolor=colour,
            **arrow_kwargs,  # type: ignore[arg-type]
        )
        arr.set_sketch_params(1, 100, 2)
        # Ignore potential overflow when using log scale. Infinity is fine.
        with np.errstate(over="ignore"):
            ax.add_patch(arr)

    # Update the axes view. ax.add_patch() doesn't do this itself.
    ax.autoscale_view()

    # Calculate an offset for the space between arrowhead and tube.
    xlim = ax.get_xlim()
    offset = 0.01 * (xlim[1] - xlim[0])

    def migration_line(source, dest, time, metadata, **mig_kwargs):
        """Draw a migration line from source to dest at the given time."""
        source_size1, source_size2 = tubes[source].sizes_at(time)
        dest_size1, dest_size2 = tubes[dest].sizes_at(time)

        if source_size2 < dest_size1:
            x = [source_size2 + offset, dest_size1 - offset]
            arrow = ">"
            lr = "r"
        else:
            x = [source_size1 - offset, dest_size2 + offset]
            arrow = "<"
            lr = "l"

        metadata = dict(lr=lr, **metadata)

        colour = colours[source]
        lines = ax.plot(
            x,
            [time, time],
            color=colour,
            path_effects=[matplotlib.patheffects.Normal()],
            zorder=-1,
            **mig_kwargs,
        )
        for line in lines:
            line.set_sketch_params(1, 100, 2)
            line._demesdraw_data = metadata
        lines = ax.plot(
            x[1],
            time,
            arrow,
            markersize=2,
            color=colour,
            path_effects=[matplotlib.patheffects.Normal()],
            **mig_kwargs,
        )
        for line in lines:
            line._demesdraw_data = metadata

    def random_migration_time(migration, log_scale: bool) -> float:
        start_time = migration.start_time
        if np.isinf(start_time):
            start_time = inf_start_time
        if log_scale:
            t = np.exp(
                rng.uniform(np.log(max(1, migration.end_time)), np.log(start_time))
            )
        else:
            t = rng.uniform(migration.end_time, start_time)
        return t

    # Plot migration lines.
    migration_kwargs = dict(linewidth=0.2, alpha=0.5)
    for migration in graph.migrations:
        for _ in range(num_lines_per_migration):
            t = random_migration_time(migration, log_time)
            migration_line(
                migration.source,
                migration.dest,
                t,
                dict(migration=migration),
                **migration_kwargs,
            )

    # Plot pulse lines.
    for pulse in graph.pulses:
        for j, source in enumerate(pulse.sources):
            migration_line(
                source,
                pulse.dest,
                pulse.time,
                dict(pulse=pulse, source=j),
                linestyle="--",
                linewidth=1,
            )

    xticks = []
    xticklabels = []
    deme_labels = [deme.name for deme in graph.demes]

    if labels in ("xticks", "xticks-legend", "xticks-mid"):
        # Put labels for the leaf nodes underneath the figure.
        leaves = [deme.name for deme in graph.demes if deme.end_time == 0]
        deme_labels = [deme.name for deme in graph.demes if deme.end_time != 0]
        xticks = [positions[leaf] for leaf in leaves]
        xticklabels = leaves

    if len(deme_labels) > 0 and labels in ("legend", "xticks-legend"):
        # Add legend.
        ax.legend(
            handles=[
                matplotlib.patches.Patch(
                    edgecolor=matplotlib.colors.to_rgba(colours[deme_name], 1.0),
                    facecolor=(
                        matplotlib.colors.to_rgba(colours[deme_name], 0.3)
                        if fill
                        else None
                    ),
                    label=deme_name,
                )
                for deme_name in deme_labels
            ],
            # Use a horizontal layout, rather than vertical.
            ncol=max(1, len(deme_labels) // 2),
        )

    if labels in ("mid", "xticks-mid"):
        for deme_name in deme_labels:
            if log_time:
                tmid = np.exp(
                    (
                        np.log(tubes[deme_name].time[0])
                        + np.log(max(1, tubes[deme_name].time[-1]))
                    )
                    / 2
                )
            else:
                tmid = (tubes[deme_name].time[0] + tubes[deme_name].time[-1]) / 2
            ax.text(
                positions[deme_name],
                tmid,
                deme_name,
                ha="center",
                va="center",
                # Give the text some contrast with its background.
                bbox=dict(boxstyle="round", fc="white", ec="none", alpha=0.6, pad=0.2),
            )

    if title is not None:
        ax.set_title(title)

    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels)
    ax.tick_params(axis="x", length=0)
    ax.tick_params(axis="x", which="minor", bottom=False)
    ax.spines[["top", "right", "bottom"]].set_visible(False)

    ax.set_ylabel(f"time ago ({graph.time_units})")

    ax.set_ylim(1 if log_time else 0, inf_start_time)

    if scale_bar:
        # Plot a scale bar that indicates the population size.
        width = utils.size_max(graph)
        transform = matplotlib.transforms.blended_transform_factory(
            ax.transData, ax.transAxes
        )
        x_offset = min(min(t.size1) for t in tubes.values())
        ax_sb = ax.inset_axes((x_offset, -0.16, width, 1e-6), transform=transform)

        ax_sb.yaxis.set_major_locator(matplotlib.ticker.NullLocator())
        ax_sb.spines[["left", "right", "top"]].set_visible(False)
        ax_sb.xaxis.set_ticks_position("top")
        locator = matplotlib.ticker.AutoLocator()
        locator.set_params(integer=True)
        ax_sb.xaxis.set_major_locator(locator)
        ax_sb.xaxis.set_minor_locator(matplotlib.ticker.AutoMinorLocator())
        ax_sb.set_xlim(0, width)
        ax_sb.set_xlabel("deme size (individuals)")
        ax_sb.xaxis.set_label_position("bottom")

    # Status bar text when in interactive mode.
    def format_coord(x, y):
        return f"time={y:.1f}"

    ax.format_coord = format_coord  # type: ignore[method-assign]

    # Note: we must hold a reference to the motion_notify_event callback,
    # or the callback will get garbage collected and thus disabled.
    # So we just add an attribute to the returned Axes object.
    ax._demesdraw_mouseoverpopup = (  # type: ignore[attr-defined]
        _MouseoverPopup(ax.figure, ax)
    )

    return ax


# Copied and modified from:
# https://matplotlib.org/stable/tutorials/advanced/blitting.html
class _BlitManager:
    def __init__(self, canvas, animated_artists=()):
        """
        Parameters
        ----------
        canvas : FigureCanvasAgg
            The canvas to work with, this only works for sub-classes of the Agg
            canvas which have the `~FigureCanvasAgg.copy_from_bbox` and
            `~FigureCanvasAgg.restore_region` methods.

        animated_artists : Iterable[Artist]
            List of the artists to manage
        """
        self.canvas = canvas
        self._bg = None
        self._artists = []

        if hasattr(canvas, "copy_from_bbox") and hasattr(canvas, "restore_region"):
            # Webagg and related backends support blitting as of matplotlib 3.4,
            # but supports_blit is False because of some rendering artifacts.
            # However, not using blitting is very laggy, so we use it anyway.
            self.canvas.supports_blit = True
        if not self.canvas.supports_blit:
            return

        for a in animated_artists:
            self.add_artist(a)
        # grab the background on every draw
        self.cid = canvas.mpl_connect("draw_event", self.on_draw)

    def on_draw(self, event):
        """Callback to register with 'draw_event'."""
        cv = self.canvas
        if event is not None:
            if event.canvas != cv:
                return
        self._bg = cv.copy_from_bbox(cv.figure.bbox)
        self._draw_animated()

    def add_artist(self, art):
        """
        Add an artist to be managed.

        Parameters
        ----------
        art : Artist

            The artist to be added.  Will be set to 'animated' (just
            to be safe).  *art* must be in the figure associated with
            the canvas this class is managing.

        """
        assert art.figure == self.canvas.figure
        art.set_animated(True)
        self._artists.append(art)

    def _draw_animated(self):
        """Draw all of the animated artists."""
        fig = self.canvas.figure
        for a in self._artists:
            fig.draw_artist(a)

    def update(self):
        """Update the screen with animated artists."""
        if not self.canvas.supports_blit:
            self.canvas.draw_idle()
            return
        cv = self.canvas
        fig = cv.figure
        # paranoia in case we missed the draw event,
        if self._bg is None:
            self.on_draw(None)
        else:
            # restore the background
            cv.restore_region(self._bg)
            # draw all of the animated artists
            self._draw_animated()
            # update the GUI state
            cv.blit(fig.bbox)


class _MouseoverPopup:
    def __init__(self, fig, ax):
        """Setup interactive mouseover annotations."""

        self.ax = ax
        self.popup = ax.annotate(
            "",
            xy=(0, 0),
            xytext=(20, 20),
            textcoords="offset points",
            multialignment="left",
            horizontalalignment="center",
            verticalalignment="center",
            bbox=dict(boxstyle="round", fc="yellow"),
            arrowprops=dict(arrowstyle="->"),
        )
        self.popup.set_visible(False)
        self.bm = _BlitManager(fig.canvas, [self.popup])

        # Detect mouseover for artists with a _demesdraw_data attribute,
        # which were added in demesdraw.tubes().
        self.artists = []
        for artist in reversed(ax.get_children()):
            if hasattr(artist, "_demesdraw_data"):
                self.artists.append(artist)

        # Get the midpoint of the figure in data coords.
        axis_to_data = self.ax.transAxes + self.ax.transData.inverted()
        self.xmid, self.ymid = axis_to_data.transform((0.5, 0.5))

        # Save a reference to the callback so that it doesn't get garbage collected.
        self.cid = fig.canvas.mpl_connect("motion_notify_event", self.motion_callback)

    def motion_callback(self, event):
        visible = self.popup.get_visible()
        mouse_inside_artist = False
        if event.inaxes == self.ax:
            for artist in self.artists:
                contained, _ = artist.contains(event)
                if contained:
                    data = artist._demesdraw_data
                    label = self.get_mouseover_label(data, event)
                    x = event.xdata
                    y = event.ydata
                    xt = 50
                    yt = 50
                    if x > self.xmid:
                        xt *= -1
                    if y > self.ymid:
                        yt *= -1
                    self.popup.xy = (x, y)
                    self.popup.xyann = (xt, yt)  # xytext
                    self.popup.set_text(label)
                    self.popup.set_visible(True)
                    mouse_inside_artist = True
                    self.bm.update()
                    break

        if not mouse_inside_artist and visible:
            self.popup.set_visible(False)
            self.bm.update()

    def get_arrow(self, lr, source, dest):
        assert lr in ("l", "r")
        arrow = r"$\rightarrow$" if lr == "r" else r"$\leftarrow$"
        if lr == "r":
            return f"{source} {arrow} {dest}"
        else:
            return f"{dest} {arrow} {source}"

    def get_mouseover_label(self, data, event):
        if "deme" in data:
            deme = data["deme"]
            time = max(min(event.ydata, deme.start_time - 1e-6), deme.end_time)
            size = deme.size_at(time)
            ancestors = ", ".join(deme.ancestors)
            proportions = ", ".join(map(str, deme.proportions))
            label = (
                f"deme: {deme.name}\n"
                f"time: ({deme.start_time}, {deme.end_time}]\n"
                f"ancestors: {ancestors}\n"
                f"proportions: {proportions}\n"
                f"size: {int(size)}"
            )
        elif "migration" in data:
            migration = data["migration"]
            arrow = self.get_arrow(data["lr"], migration.source, migration.dest)
            label = (
                f"migration: {arrow}\n"
                f"time: ({migration.start_time}, {migration.end_time}]\n"
                f"rate: {migration.rate}"
            )
        elif "pulse" in data:
            pulse = data["pulse"]
            j = data["source"]
            arrow = self.get_arrow(data["lr"], pulse.sources[j], pulse.dest)
            label = (
                f"pulse: {arrow}\n"
                f"time: {pulse.time}\n"
                f"proportion: {pulse.proportions[j]}"
            )
        return label



from __future__ import annotations

import demes
import numpy as np
import matplotlib

from demesdraw import utils


def size_history(
    graph: demes.Graph,
    ax: matplotlib.axes.Axes | None = None,
    colours: utils.ColourOrColourMapping | None = None,
    log_time: bool = False,
    log_size: bool = False,
    title: str | None = None,
    inf_ratio: float = 0.1,
    inf_label: bool = False,
    invert_x: bool = False,
    annotate_epochs: bool = False,
    num_points: int = 100,
) -> matplotlib.axes.Axes:
    """
    Plot population size as a function of time for each deme in the graph.

    :param demes.Graph graph:
        The demes graph to plot.
    :param ax:
        The matplotlib axes onto which the figure
        will be drawn. If None, an empty axes will be created for the figure.
    :type ax: Optional[matplotlib.axes.Axes]
    :param colours:
        A mapping from deme name to matplotlib colour. Alternately,
        ``colours`` may be a named colour that will be used for all demes.
    :type colours: Optional[dict or str]
    :param log_time:
        If True, use a log-10 scale for the time axis.
        If False (*default*), a linear scale will be used.
    :param bool log_size:
        If True, use a log-10 scale for the size axis.
        If False (*default*), a linear scale will be used.
    :param title:
        The title of the figure.
    :param inf_ratio:
        The proportion of the time axis that will be
        used for the time interval which stretches towards infinity.
    :param inf_label:
        Write "inf" by the arrow that points towards infinity.
    :param invert_x:
        If True, the horizontal axis will have infinity on the left and
        zero on the right, and the vertical axis will be drawn on the right.
        If False (*default*), the horizontal axis will have zero on the left
        and infinity on the right, and the vertical axis will be drawn
        on the left.
    :param annotate_epochs:
        If True, annotate the figure with epoch indices over the relevant
        parts of the lines. This may be useful as a pedagogical tool.
        If False (*default*), do not annotate the epochs.
    :return: The matplotlib axes onto which the figure was drawn.
    """
    if ax is None:
        _, ax = utils.get_fig_axes()

    if invert_x:
        arrowhead = "<"
    else:
        arrowhead = ">"

    colours = utils._get_colours(graph, colours)
    inf_start_time = utils._inf_start_time(graph, inf_ratio, log_time)

    linestyles = ["solid"]  # , "dashed", "dashdot"]
    linewidths = [2, 4, 8, 1]
    legend_handles = []
    # Top of the z order stacking.
    z_top = 1 + len(graph.demes) + max(linewidths)

    for j, deme in enumerate(graph.demes):
        colour = colours[deme.name]
        linestyle = linestyles[j % len(linestyles)]
        linewidth = linewidths[j % len(linewidths)]
        plot_kwargs = dict(
            color=colour,
            linestyle=linestyle,
            linewidth=linewidth,
            label=deme.name,
            alpha=0.7,
            zorder=z_top - linewidth,
            capstyle="butt",
            fill=False,
            path_effects=[
                matplotlib.patheffects.withStroke(linewidth=3, foreground="white")
            ],
        )
        discontinuity_kwargs = plot_kwargs.copy()
        discontinuity_kwargs.update(linestyle=":")
        legend_kwargs = plot_kwargs.copy()
        legend_kwargs.pop("fill")
        # Line2D and Patch use different keywords for the capstyle. *Sigh*
        legend_kwargs.update(solid_capstyle=legend_kwargs.pop("capstyle"))

        legend_handles.append(
            matplotlib.lines.Line2D([], [], **legend_kwargs)  # type: ignore[arg-type]
        )

        # Path for the main line (solid).
        vertices_main = []
        codes_main = []
        # Path for the discontinuity lines (dashed).
        vertices_discontinuity = []
        codes_discontinuity = []

        for k, epoch in enumerate(deme.epochs):
            start_time = epoch.start_time
            if np.isinf(start_time):
                start_time = inf_start_time
            end_time = epoch.end_time
            if log_time:
                end_time = max(1, end_time)

            if epoch.size_function == "constant":
                x = np.array([start_time, end_time])
                y = np.array([epoch.start_size, epoch.end_size])
            elif epoch.size_function == "exponential":
                x = np.linspace(start_time, end_time, num=num_points)
                dt = np.linspace(0, 1, num=num_points)
                r = np.log(epoch.end_size / epoch.start_size)
                y = epoch.start_size * np.exp(r * dt)
            elif epoch.size_function == "linear":
                x = np.linspace(start_time, end_time, num=num_points)
                dt = np.linspace(0, 1, num=num_points)
                y = epoch.start_size + (epoch.end_size - epoch.start_size) * dt
            else:
                raise ValueError(
                    f"Don't know how to plot epoch {k} with "
                    f'"{epoch.size_function}" size_function.'
                )

            vertices_main.extend(list(zip(x, y)))
            if k == 0 or deme.epochs[k - 1].end_size != epoch.start_size:
                codes_main.append(matplotlib.path.Path.MOVETO)
            else:
                codes_main.append(matplotlib.path.Path.LINETO)
            codes_main.extend([matplotlib.path.Path.LINETO] * (len(x) - 1))

            if k > 0 and deme.epochs[k - 1].end_size != epoch.start_size:
                # Size discontinuity.
                vertices_discontinuity.extend(
                    [
                        (deme.epochs[k - 1].end_time, deme.epochs[k - 1].end_size),
                        (epoch.start_time, epoch.start_size),
                    ]
                )
                codes_discontinuity.extend(
                    [matplotlib.path.Path.MOVETO, matplotlib.path.Path.LINETO]
                )

            if annotate_epochs:
                if log_time:
                    text_x = np.exp((np.log(start_time) + np.log(end_time)) / 2)
                else:
                    text_x = (start_time + end_time) / 2
                if log_size:
                    text_y = np.exp(
                        (np.log(epoch.start_size) + np.log(max(1, epoch.end_size))) / 2
                    )
                else:
                    text_y = (epoch.start_size + epoch.end_size) / 2
                ax.annotate(
                    f"epoch {k}",
                    (text_x, text_y),
                    ha="center",
                    va="bottom",
                    xytext=(0, 4 + linewidth / 2),  # vertical offset
                    textcoords="offset points",
                    # Give the text some contrast with its background.
                    bbox=dict(
                        boxstyle="round", fc="white", ec="none", alpha=0.6, pad=0
                    ),
                    # This is only really a useful feature with 1 deme,
                    # but at least try to do something reasonable for more demes.
                    color="black" if len(graph.demes) == 1 else colour,
                    zorder=z_top,
                )

        # Indicate population size discontinuities from ancestor demes.
        for ancestor_id in deme.ancestors:
            anc = graph[ancestor_id]
            anc_N = anc.size_at(deme.start_time)
            deme_N = deme.epochs[0].start_size
            if anc_N != deme_N:
                vertices_discontinuity.extend(
                    [(deme.start_time, anc_N), (deme.start_time, deme_N)]
                )
                codes_discontinuity.extend(
                    [matplotlib.path.Path.MOVETO, matplotlib.path.Path.LINETO]
                )

        size_path_patch = matplotlib.patches.PathPatch(
            matplotlib.path.Path(vertices_main, codes_main), **plot_kwargs
        )
        ax.add_patch(size_path_patch)
        if len(vertices_discontinuity) > 0:
            discontinuity_path_patch = matplotlib.patches.PathPatch(
                matplotlib.path.Path(vertices_discontinuity, codes_discontinuity),
                **discontinuity_kwargs,
            )
            ax.add_patch(discontinuity_path_patch)

        if np.isinf(deme.start_time):
            # Plot an arrow at the end of the line, to indicate this
            # line extends towards infinity.
            ax.plot(
                inf_start_time,
                deme.epochs[0].start_size,
                arrowhead,
                color=colour,
                clip_on=False,
                zorder=z_top,
            )
            if inf_label:
                ax.annotate(
                    "inf",
                    (inf_start_time, deme.epochs[0].start_size),
                    xytext=(0, -6),  # vertical offset
                    textcoords="offset points",
                    clip_on=False,
                    ha="center",
                    va="top",
                )

    # Update the axes view. ax.add_patch() doesn't do this itself.
    ax.autoscale_view()

    if len(graph.demes) > 1:
        leg = ax.legend(handles=legend_handles, ncol=max(1, len(graph.demes) // 2))
        leg.set_zorder(z_top)

    if title is not None:
        ax.set_title(title)

    # Arrange the axes spines, ticks and labels.

    ax.set_xlim(1 if log_time else 0, inf_start_time)
    if not log_size:
        ax.set_ylim(bottom=0)

    for spine in ax.spines.values():
        spine.set_zorder(z_top)

    ax.spines["top"].set_visible(False)
    if invert_x:
        ax.spines["left"].set_visible(False)
        ax.invert_xaxis()
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position("right")
    else:
        ax.spines["right"].set_visible(False)

    ax.set_xlabel(f"time ago ({graph.time_units})")
    # ax.set_ylabel("N", rotation=0, ha="left" if invert_x else "right")
    ax.set_ylabel("deme\nsize", rotation=0, labelpad=20)

    if log_time:
        ax.set_xscale("log", base=10)
    if log_size:
        ax.set_yscale("log", base=10)

    return ax
